@@ -7,6 +7,8 @@ use minicbor::{bytes::ByteVec, data::Tag};
 use minicbor_derive::{Decode, Encode};
 use std::{collections::BTreeMap, ops::Deref};
 
+use crate::utils::{KeyValuePairs, MaybeIndefArray};
+
 #[derive(Debug, PartialEq, PartialOrd, Eq, Ord)]
 pub struct SkipCbor<const N: usize> {}
 
@@ -112,7 +114,7 @@ pub type PolicyId = ScriptHash;
 
 pub type AssetName = ByteVec;
 
-pub type Multiasset<A> = BTreeMap<PolicyId, BTreeMap<AssetName, A>>;
+pub type Multiasset<A> = KeyValuePairs<PolicyId, KeyValuePairs<AssetName, A>>;
 
 pub type Mint = Multiasset<i64>;
 
@@ -612,136 +614,148 @@ pub enum NetworkId {
     Two,
 }
 
-#[derive(Decode, Debug, PartialEq)]
-#[cbor(map)]
-pub struct TransactionBody {
-    #[n(0)]
-    pub inputs: Vec<TransactionInput>,
+#[derive(Debug, PartialEq)]
+pub enum TransactionBodyComponent {
+    Inputs(MaybeIndefArray<TransactionInput>),
+    Outputs(MaybeIndefArray<TransactionOutput>),
+    Fee(u64),
+    Ttl(u64),
+    Certificates(Vec<Certificate>),
+    Withdrawals(KeyValuePairs<RewardAccount, Coin>),
+    Update(Option<SkipCbor<22>>),
+    AuxiliaryDataHash(ByteVec),
+    ValidityIntervalStart(u64),
+    Mint(Multiasset<i64>),
+    ScriptDataHash(Hash32),
+    Collateral(MaybeIndefArray<TransactionInput>),
+    RequiredSigners(MaybeIndefArray<AddrKeyhash>),
+    NetworkId(NetworkId),
+}
+impl<'b> minicbor::decode::Decode<'b> for TransactionBodyComponent {
+    fn decode(d: &mut minicbor::Decoder<'b>) -> Result<Self, minicbor::decode::Error> {
+        let key: u32 = d.decode()?;
 
-    #[n(1)]
-    pub outputs: Vec<TransactionOutput>,
+        match key {
+            0 => Ok(Self::Inputs(d.decode()?)),
+            1 => Ok(Self::Outputs(d.decode()?)),
+            2 => Ok(Self::Fee(d.decode()?)),
+            3 => Ok(Self::Ttl(d.decode()?)),
+            4 => Ok(Self::Certificates(d.decode()?)),
+            5 => Ok(Self::Withdrawals(d.decode()?)),
+            6 => Ok(Self::Update(d.decode()?)),
+            7 => Ok(Self::AuxiliaryDataHash(d.decode()?)),
+            8 => Ok(Self::ValidityIntervalStart(d.decode()?)),
+            9 => Ok(Self::Mint(d.decode()?)),
+            11 => Ok(Self::ScriptDataHash(d.decode()?)),
+            13 => Ok(Self::Collateral(d.decode()?)),
+            14 => Ok(Self::RequiredSigners(d.decode()?)),
+            15 => Ok(Self::NetworkId(d.decode()?)),
+            _ => Err(minicbor::decode::Error::Message(
+                "invalid map key for transaction body component",
+            )),
+        }
+    }
+}
 
-    #[n(2)]
-    pub fee: u64,
+impl minicbor::encode::Encode for TransactionBodyComponent {
+    fn encode<W: minicbor::encode::Write>(
+        &self,
+        e: &mut minicbor::Encoder<W>,
+    ) -> Result<(), minicbor::encode::Error<W::Error>> {
+        match self {
+            TransactionBodyComponent::Inputs(x) => {
+                e.encode(0)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Outputs(x) => {
+                e.encode(1)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Fee(x) => {
+                e.encode(2)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Ttl(x) => {
+                e.encode(3)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Certificates(x) => {
+                e.encode(4)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Withdrawals(x) => {
+                e.encode(5)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Update(x) => {
+                e.encode(6)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::AuxiliaryDataHash(x) => {
+                e.encode(7)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::ValidityIntervalStart(x) => {
+                e.encode(8)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Mint(x) => {
+                e.encode(9)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::ScriptDataHash(x) => {
+                e.encode(11)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::Collateral(x) => {
+                e.encode(13)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::RequiredSigners(x) => {
+                e.encode(14)?;
+                e.encode(x)?;
+            }
+            TransactionBodyComponent::NetworkId(x) => {
+                e.encode(15)?;
+                e.encode(x)?;
+            }
+        }
 
-    #[n(3)]
-    pub ttl: Option<u64>,
-
-    #[n(4)]
-    pub certificates: Option<Vec<Certificate>>,
-
-    #[n(5)]
-    pub withdrawals: Option<BTreeMap<RewardAccount, Coin>>,
-
-    #[n(6)]
-    pub update: Option<SkipCbor<22>>,
-
-    #[n(7)]
-    pub auxiliary_data_hash: Option<ByteVec>,
-
-    #[n(8)]
-    pub validity_interval_start: Option<u64>,
-
-    #[n(9)]
-    pub mint: Option<Multiasset<i64>>,
-
-    #[n(11)]
-    pub script_data_hash: Option<Hash32>,
-
-    #[n(13)]
-    pub collateral: Option<Vec<TransactionInput>>,
-
-    #[n(14)]
-    pub required_signers: Option<Vec<AddrKeyhash>>,
-
-    #[n(15)]
-    pub network_id: Option<NetworkId>,
+        Ok(())
+    }
 }
 
 // Can't derive encode for TransactionBody because it seems to require a very
 // particular order for each key in the map
+#[derive(Debug, PartialEq)]
+pub struct TransactionBody(Vec<TransactionBodyComponent>);
+
+impl Deref for TransactionBody {
+    type Target = Vec<TransactionBodyComponent>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl<'b> minicbor::decode::Decode<'b> for TransactionBody {
+    fn decode(d: &mut minicbor::Decoder<'b>) -> Result<Self, minicbor::decode::Error> {
+        let len = d.map()?.unwrap_or_default();
+
+        let components: Result<_, _> = (0..len).map(|_| d.decode()).collect();
+
+        Ok(Self(components?))
+    }
+}
+
 impl minicbor::encode::Encode for TransactionBody {
     fn encode<W: minicbor::encode::Write>(
         &self,
         e: &mut minicbor::Encoder<W>,
     ) -> Result<(), minicbor::encode::Error<W::Error>> {
-        let map_size = 3
-            + self.ttl.map_or(0, |_| 1)
-            + self.certificates.as_ref().map_or(0, |_| 1)
-            + self.withdrawals.as_ref().map_or(0, |_| 1)
-            + self.update.as_ref().map_or(0, |_| 1)
-            + self.auxiliary_data_hash.as_deref().map_or(0, |_| 1)
-            + self.validity_interval_start.map_or(0, |_| 1)
-            + self.mint.as_ref().map_or(0, |_| 1)
-            + self.script_data_hash.as_deref().map_or(0, |_| 1)
-            + self.collateral.as_deref().map_or(0, |_| 1)
-            + self.required_signers.as_deref().map_or(0, |_| 1)
-            + self.network_id.as_ref().map_or(0, |_| 1);
-
-        e.map(map_size)?;
-
-        e.encode(0)?;
-        e.encode(&self.inputs)?;
-
-        if let Some(x) = &self.collateral {
-            e.encode(13)?;
-            e.encode(x)?;
-        }
-
-        e.encode(1)?;
-        e.encode(&self.outputs)?;
-
-        e.encode(2)?;
-        e.encode(&self.fee)?;
-
-        if let Some(x) = &self.ttl {
-            e.encode(3)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.certificates {
-            e.encode(4)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.withdrawals {
-            e.encode(5)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.update {
-            e.encode(6)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.validity_interval_start {
-            e.encode(8)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.required_signers {
-            e.encode(14)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.script_data_hash {
-            e.encode(11)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.network_id {
-            e.encode(15)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.auxiliary_data_hash {
-            e.encode(7)?;
-            e.encode(x)?;
-        }
-
-        if let Some(x) = &self.mint {
-            e.encode(9)?;
-            e.encode(x)?;
+        e.map(self.0.len() as u64)?;
+        for component in &self.0 {
+            e.encode(component)?;
         }
 
         Ok(())
@@ -760,9 +774,9 @@ pub struct VKeyWitness {
 #[derive(Debug, PartialEq)]
 pub enum NativeScript {
     ScriptPubkey(AddrKeyhash),
-    ScriptAll(Vec<NativeScript>),
-    ScriptAny(Vec<NativeScript>),
-    ScriptNOfK(u32, Vec<NativeScript>),
+    ScriptAll(MaybeIndefArray<NativeScript>),
+    ScriptAny(MaybeIndefArray<NativeScript>),
+    ScriptNOfK(u32, MaybeIndefArray<NativeScript>),
     InvalidBefore(u64),
     InvalidHereafter(u64),
 }
@@ -827,14 +841,81 @@ impl minicbor::encode::Encode for NativeScript {
 
 pub type PlutusScript = ByteVec;
 
+/*
+big_int = int / big_uint / big_nint ; New
+big_uint = #6.2(bounded_bytes) ; New
+big_nint = #6.3(bounded_bytes) ; New
+ */
+
+#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+pub enum BigInt {
+    Int(i64),
+    BigUInt(ByteVec),
+    BigNInt(ByteVec),
+}
+
+impl<'b> minicbor::decode::Decode<'b> for BigInt {
+    fn decode(d: &mut minicbor::Decoder<'b>) -> Result<Self, minicbor::decode::Error> {
+        let datatype = d.datatype()?;
+
+        match datatype {
+            minicbor::data::Type::U8
+            | minicbor::data::Type::U16
+            | minicbor::data::Type::U32
+            | minicbor::data::Type::U64
+            | minicbor::data::Type::I8
+            | minicbor::data::Type::I16
+            | minicbor::data::Type::I32
+            | minicbor::data::Type::I64 => Ok(Self::Int(d.decode()?)),
+            minicbor::data::Type::Tag => {
+                let tag = d.tag()?;
+
+                match tag {
+                    minicbor::data::Tag::PosBignum => Ok(Self::BigUInt(d.decode()?)),
+                    minicbor::data::Tag::NegBignum => Ok(Self::BigNInt(d.decode()?)),
+                    _ => Err(minicbor::decode::Error::Message(
+                        "invalid cbor tag for big int",
+                    )),
+                }
+            }
+            _ => Err(minicbor::decode::Error::Message(
+                "invalid cbor data type for big int",
+            )),
+        }
+    }
+}
+
+impl minicbor::encode::Encode for BigInt {
+    fn encode<W: minicbor::encode::Write>(
+        &self,
+        e: &mut minicbor::Encoder<W>,
+    ) -> Result<(), minicbor::encode::Error<W::Error>> {
+        match self {
+            BigInt::Int(x) => {
+                e.encode(x)?;
+            }
+            BigInt::BigUInt(x) => {
+                e.tag(Tag::PosBignum)?;
+                e.encode(x)?;
+            }
+            BigInt::BigNInt(x) => {
+                e.tag(Tag::NegBignum)?;
+                e.encode(x)?;
+            }
+        };
+
+        Ok(())
+    }
+}
+
 #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
 pub enum PlutusData {
     Constr(Constr<PlutusData>),
-    Map(BTreeMap<PlutusData, PlutusData>),
-    BitInt(u64),
+    Map(KeyValuePairs<PlutusData, PlutusData>),
+    BigInt(BigInt),
     BoundedBytes(ByteVec),
-    Array(Vec<PlutusData>),
-    ArrayIndef(IndefVec<PlutusData>),
+    Array(MaybeIndefArray<PlutusData>),
+    ArrayIndef(MaybeIndefArray<PlutusData>),
 }
 
 impl<'b> minicbor::decode::Decode<'b> for PlutusData {
@@ -842,19 +923,32 @@ impl<'b> minicbor::decode::Decode<'b> for PlutusData {
         let type_ = d.datatype()?;
 
         match type_ {
-            minicbor::data::Type::Tag => Ok(PlutusData::Constr(d.decode()?)),
-            minicbor::data::Type::Map => Ok(PlutusData::Map(d.decode()?)),
-            minicbor::data::Type::I8 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::I16 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::I32 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::I64 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::U8 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::U16 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::U32 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::U64 => Ok(PlutusData::BitInt(d.decode()?)),
-            minicbor::data::Type::Bytes => Ok(PlutusData::BoundedBytes(d.decode()?)),
-            minicbor::data::Type::Array => Ok(PlutusData::Array(d.decode()?)),
-            minicbor::data::Type::ArrayIndef => Ok(PlutusData::ArrayIndef(d.decode()?)),
+            minicbor::data::Type::Tag => {
+                let mut probe = d.probe();
+                let tag = probe.tag()?;
+
+                match tag {
+                    Tag::Unassigned(121..=127 | 1280..=1400 | 102) => Ok(Self::Constr(d.decode()?)),
+                    Tag::PosBignum | Tag::NegBignum => Ok(Self::BigInt(d.decode()?)),
+                    _ => Err(minicbor::decode::Error::Message(
+                        "unknown tag for plutus data tag",
+                    )),
+                }
+            }
+            minicbor::data::Type::U8
+            | minicbor::data::Type::U16
+            | minicbor::data::Type::U32
+            | minicbor::data::Type::U64
+            | minicbor::data::Type::I8
+            | minicbor::data::Type::I16
+            | minicbor::data::Type::I32
+            | minicbor::data::Type::I64 => Ok(Self::BigInt(d.decode()?)),
+            minicbor::data::Type::Map => Ok(Self::Map(d.decode()?)),
+            minicbor::data::Type::Bytes => Ok(Self::BoundedBytes(d.decode()?)),
+            minicbor::data::Type::BytesIndef => Ok(Self::BoundedBytes(d.decode()?)),
+            minicbor::data::Type::Array => Ok(Self::Array(d.decode()?)),
+            minicbor::data::Type::ArrayIndef => Ok(Self::ArrayIndef(d.decode()?)),
+
             _ => Err(minicbor::decode::Error::Message(
                 "bad cbor data type for plutus data",
             )),
@@ -868,66 +962,25 @@ impl minicbor::encode::Encode for PlutusData {
         e: &mut minicbor::Encoder<W>,
     ) -> Result<(), minicbor::encode::Error<W::Error>> {
         match self {
-            PlutusData::Constr(a) => {
+            Self::Constr(a) => {
                 e.encode(a)?;
-                Ok(())
             }
-            PlutusData::Map(a) => {
+            Self::Map(a) => {
                 e.encode(a)?;
-                Ok(())
             }
-            PlutusData::BitInt(a) => {
+            Self::BigInt(a) => {
                 e.encode(a)?;
-                Ok(())
             }
-            PlutusData::BoundedBytes(a) => {
+            Self::BoundedBytes(a) => {
                 e.encode(a)?;
-                Ok(())
             }
-            PlutusData::Array(a) => {
+            Self::Array(a) => {
                 e.encode(a)?;
-                Ok(())
             }
-            PlutusData::ArrayIndef(a) => {
+            Self::ArrayIndef(a) => {
                 e.encode(a)?;
-                Ok(())
             }
         }
-    }
-}
-
-/// A struct that forces encode / decode using indef arrays
-#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
-pub struct IndefVec<A>(pub Vec<A>);
-
-impl<'b, A> minicbor::decode::Decode<'b> for IndefVec<A>
-where
-    A: minicbor::decode::Decode<'b>,
-{
-    fn decode(d: &mut minicbor::Decoder<'b>) -> Result<Self, minicbor::decode::Error> {
-        let values: Vec<A> = d.decode()?;
-
-        Ok(IndefVec(values))
-    }
-}
-
-impl<A> minicbor::encode::Encode for IndefVec<A>
-where
-    A: minicbor::encode::Encode,
-{
-    fn encode<W: minicbor::encode::Write>(
-        &self,
-        e: &mut minicbor::Encoder<W>,
-    ) -> Result<(), minicbor::encode::Error<W::Error>> {
-        if self.0.len() > 0 {
-            e.begin_array()?;
-            for v in &self.0 {
-                e.encode(v)?;
-            }
-            e.end()?;
-        } else {
-            e.array(0)?;
-        }
 
         Ok(())
     }
@@ -937,7 +990,7 @@ where
 pub struct Constr<A> {
     pub tag: u64,
     pub prefix: Option<u32>,
-    pub values: IndefVec<A>,
+    pub values: MaybeIndefArray<A>,
 }
 
 impl<'b, A> minicbor::decode::Decode<'b> for Constr<A>
@@ -949,7 +1002,7 @@ where
 
         match tag {
             Tag::Unassigned(x) => match x {
-                121 | 122 | 123 | 124 | 125 | 126 | 127 => Ok(Constr {
+                121..=127 | 1280..=1400 => Ok(Constr {
                     tag: x,
                     values: d.decode()?,
                     prefix: None,
@@ -1064,22 +1117,22 @@ pub struct BootstrapWitness {
 #[cbor(map)]
 pub struct TransactionWitnessSet {
     #[n(0)]
-    pub vkeywitness: Option<Vec<VKeyWitness>>,
+    pub vkeywitness: Option<MaybeIndefArray<VKeyWitness>>,
 
     #[n(1)]
-    pub native_script: Option<Vec<NativeScript>>,
+    pub native_script: Option<MaybeIndefArray<NativeScript>>,
 
     #[n(2)]
-    pub bootstrap_witness: Option<Vec<BootstrapWitness>>,
+    pub bootstrap_witness: Option<MaybeIndefArray<BootstrapWitness>>,
 
     #[n(3)]
-    pub plutus_script: Option<Vec<PlutusScript>>,
+    pub plutus_script: Option<MaybeIndefArray<PlutusScript>>,
 
     #[n(4)]
-    pub plutus_data: Option<Vec<PlutusData>>,
+    pub plutus_data: Option<MaybeIndefArray<PlutusData>>,
 
     #[n(5)]
-    pub redeemer: Option<Vec<Redeemer>>,
+    pub redeemer: Option<MaybeIndefArray<Redeemer>>,
 }
 
 #[derive(Encode, Decode, Debug, PartialEq)]
@@ -1088,7 +1141,7 @@ pub struct AlonzoAuxiliaryData {
     #[n(0)]
     pub metadata: Option<Metadata>,
     #[n(1)]
-    pub native_scripts: Option<Vec<NativeScript>>,
+    pub native_scripts: Option<MaybeIndefArray<NativeScript>>,
     #[n(2)]
     pub plutus_scripts: Option<PlutusScript>,
 }
@@ -1098,7 +1151,7 @@ pub enum Metadatum {
     Int(i64),
     Bytes(ByteVec),
     Text(String),
-    Array(Vec<Metadatum>),
+    Array(MaybeIndefArray<Metadatum>),
     Map(Metadata),
 }
 
@@ -1175,45 +1228,7 @@ impl minicbor::Encode for Metadatum {
     }
 }
 
-/// Custom struct to handle nested metadata maps
-///
-/// Since the ordering of the entires requires a particular order to maintain
-/// canonicalization for isomorphic decoding / encoding operators, we use a Vec
-/// as the underlaying struct for storage of the items (as opposed to a BTreeMap
-/// or HashMap).
-#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
-pub struct Metadata(pub Vec<(Metadatum, Metadatum)>);
-
-impl Deref for Metadata {
-    type Target = Vec<(Metadatum, Metadatum)>;
-
-    fn deref(&self) -> &Self::Target {
-        &self.0
-    }
-}
-
-impl<'b> minicbor::decode::Decode<'b> for Metadata {
-    fn decode(d: &mut minicbor::Decoder<'b>) -> Result<Self, minicbor::decode::Error> {
-        let items: Result<Vec<_>, _> = d.map_iter::<Metadatum, Metadatum>()?.collect();
-        let items = items?;
-        Ok(Metadata(items))
-    }
-}
-
-impl minicbor::encode::Encode for Metadata {
-    fn encode<W: minicbor::encode::Write>(
-        &self,
-        e: &mut minicbor::Encoder<W>,
-    ) -> Result<(), minicbor::encode::Error<W::Error>> {
-        e.map(self.0.len() as u64)?;
-        for (k, v) in &self.0 {
-            k.encode(e)?;
-            v.encode(e)?;
-        }
-
-        Ok(())
-    }
-}
+pub type Metadata = KeyValuePairs<Metadatum, Metadatum>;
 
 #[derive(Debug, PartialEq)]
 pub enum AuxiliaryData {
@@ -1285,16 +1300,16 @@ pub struct Block {
     pub header: Header,
 
     #[n(1)]
-    pub transaction_bodies: Vec<TransactionBody>,
+    pub transaction_bodies: MaybeIndefArray<TransactionBody>,
 
     #[n(2)]
-    pub transaction_witness_sets: Vec<TransactionWitnessSet>,
+    pub transaction_witness_sets: MaybeIndefArray<TransactionWitnessSet>,
 
     #[n(3)]
-    pub auxiliary_data_set: BTreeMap<TransactionIndex, AuxiliaryData>,
+    pub auxiliary_data_set: KeyValuePairs<TransactionIndex, AuxiliaryData>,
 
     #[n(4)]
-    pub invalid_transactions: Vec<TransactionIndex>,
+    pub invalid_transactions: MaybeIndefArray<TransactionIndex>,
 }
 
 #[derive(Encode, Decode, Debug)]
@@ -1316,15 +1331,17 @@ mod tests {
             include_str!("test_data/test6.block"),
             include_str!("test_data/test7.block"),
             include_str!("test_data/test8.block"),
+            include_str!("test_data/test9.block"),
         ];
 
         for (idx, block_str) in test_blocks.iter().enumerate() {
-            //println!("decoding test block {}", idx + 1);
+            println!("decoding test block {}", idx + 1);
             let bytes = hex::decode(block_str).expect(&format!("bad block file {}", idx));
             let block = BlockWrapper::decode_fragment(&bytes[..])
                 .expect(&format!("error decoding cbor for file {}", idx));
             let bytes2 =
                 to_vec(block).expect(&format!("error encoding block cbor for file {}", idx));
+
             assert_eq!(bytes, bytes2);
         }
     }
